#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <error.h>
#include <string.h>
#include <arpa/inet.h>

#include <libimon2.h>
/*
#include <libimon.h>
*/
#include <thread.h>
#include <pix.h>

#define out_info(is_out, format, args...)	if (is_out) {printf(format, ##args);}

#define SIZE_DATA_ARRAY 0x400
#define PIXEL_SIZE	sizeof(int16_t)
#define STEP_KOEF	800

#define COPY_TV(timeval_des, timeval_src)			( (timeval_des)->tv_sec=(timeval_src)->tv_sec, \
								(timeval_des)->tv_usec=(timeval_src)->tv_usec )
#define DIFF_TIME_OF_DAY(timeval_old, timeval_new)		( ((timeval_new)->tv_sec-(timeval_old)->tv_sec)*1000000 \
								+ ((timeval_new)->tv_usec-(timeval_old)->tv_usec) )

static double calc_wavelength (char *src_arr, char *first_diff_arr, char *second_diff_arr,
		size_t size_data_array, struct CalibrationCoefficient *cc, int diff_level) {
	int16_t X1, Y1, X2, Y2, index, diff;
	double calc_pixel=0;

	if (hold_array(src_arr, first_diff_arr, second_diff_arr, size_data_array,
				NULL, NULL, NULL, 0, 0, 0)<0) {
		perror("hold_array error");
		exit(4);
	}
	for (index=2; index<SIZE_DATA_ARRAY; index+=PIXEL_SIZE) {
		Y1=*((int16_t *)(second_diff_arr+index-PIXEL_SIZE));
		Y2=*((int16_t *)(second_diff_arr+index));
		diff=Y2-Y1;
		if (diff>diff_level && Y2>0 && Y1<0 ) {
			X1=(index/PIXEL_SIZE)-1, X2=index/PIXEL_SIZE;
			calc_pixel=xaxis_crossed(X1, Y1, X2, Y2);
			return wavelength_return(cc->A0, cc->B1, cc->B2, cc->B3, cc->B4, cc->B5, calc_pixel);
		}
	}
	return 0;
}

int MThread::start_capture (int fd,		/* opened driver descripto */
			double exposure_time,	/* in mcs */
			double exposure_cycle,	/* in mcs */
			int is_out,		/* print information */
			int gain,
			int datatransmit,
			int dataposition,
			int datacount,
			int count_cycles,
			int gain_mode,
			short capture_mode,
			short start_capture_mode,
			const char *out_dir) {
	struct timeval tv_begin,
		tv_end;

	int i, j, fd_out=0,
	    retval=0,
	    triggeroffset=-1,
	    trigger_mode=0x0,
	    trigger_polarity=0x0,
	    trigger_output=0x0,
	    adoffset=0x0,
	    coefs_mode=0x0;

	char temp[0x30],
	     doublestr[0x20],
	     src_array[SIZE_DATA_ARRAY],
	     first_diff_array[SIZE_DATA_ARRAY],
	     second_diff_array[SIZE_DATA_ARRAY];

	long long count_frames=0;
	long temp_long,
	     diff_time;

	double temp_double,
		wavelength=0;
	double prev_wavelength=0;

	struct ModuleInformation mi;
	struct SpectroInformation si;
	struct CalibrationCoefficient cc;
	struct ModuleProperty mp;
	struct DataPositionProperty dpp;
	struct DataTriggerOffsetProperty dtop;
	struct ExposureProperty ep;
	struct AdOffsetProperty ap;
	struct CoolingProperty cp;
	struct WavelengthProperty wp;
	struct ImageSize is;
	
	short startcaptmode=start_capture_mode; /* valid value 0x0100 */
	double poolHZ=0;
	memset(src_array, 0, SIZE_DATA_ARRAY);
	memset(first_diff_array, 0, SIZE_DATA_ARRAY);
	memset(second_diff_array, 0, SIZE_DATA_ARRAY);

	if ((imon2_get_module_information(fd, &mi))) {
		printf("error for imon2_get_module_inf\n");
	} else {
		out_info(is_out, "ModuleInformation:\n\tvender: %s\n\tproduct: %s\n\tserial: %s\n\tfirmware: %s\n",
				mi.vender, mi.product, mi.serial, mi.firmware);
	}

	if (imon2_get_spectro_information(fd, &si)) {
		printf("error for imon2_get_spectro_information\n");
	} else {
		memset(temp, 0, sizeof(temp));
		out_info(is_out,temp, 0x8, "%s", si.unit);
		out_info(is_out,"SpectroInformation:\n\tunit: %s\n\tsensor: %s\n", temp, si.sensor);
	}

	memset(temp, 0, sizeof(temp));
	if (imon2_get_calibration_coefficient(fd, coefs_mode, (struct CalibrationCoefficient*)temp)) {
		printf("error for imon2_get_calibration_coefficient\n");
	} else {
		over64((char *)&cc.A0, temp);
		over64((char *)&cc.B1, temp+8);
		over64((char *)&cc.B2, temp+16);
		over64((char *)&cc.B3, temp+24);
		over64((char *)&cc.B4, temp+32);
		over64((char *)&cc.B5, temp+40);
		out_info(is_out, "Calibration Coefficients:\n\tA0=%e\n\tB1=%e\n\tB2=%e\n\tB3=%e\n\tB4=%e\n\tB5=%e\n",
				cc.A0, cc.B1, cc.B2, cc.B3, cc.B4, cc.B5);
	}

	if (imon2_get_module_property(fd, &mp)) {
		printf("error for imon2_get_module_property\n");
	} else {
		out_info(is_out, "Module Property:\n\tpixel: %d\n\tresolution: %d"
				"\n\tcapture mode: %d\n\tbinning mode %d"
				"\n\ttrigger mode %d\n\ttrigger polarity: %d\n",
				ntohl(mp.pixel), ntohl(mp.resolution), ntohl(mp.capturemode),
				ntohl(mp.binningmode), ntohl(mp.triggermode), ntohl(mp.triggerpolarity));
	}

	if (imon2_get_dataposition_property(fd, &dpp)) {
		printf("error for imon2_get_dataposition_property\n");
	} else {
		out_info(is_out, "DataPosition Property:\n\tlower position: %d\n\tupper position: %d\n",
				ntohl(dpp.lowerposition), ntohl(dpp.upperposition));
	}

	if (imon2_get_datacount_property(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_datacount_property\n");
	} else {
		out_info(is_out, "DataCount Property:\n\t %d\n", ntohl(temp_long));
	}

	if (imon2_get_datatransmit_property(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_datatransmit_property\n");
	} else {
		out_info(is_out, "DataTransmit Property:\n\t %d\n", ntohl(temp_long));
	}

	if (imon2_get_datatriggeroffset_property(fd, &dtop)) {
		printf("error for imon2_get_datatriggeroffset_property\n");
	} else {
		out_info(is_out, "DataTriggerOffset Property:\n\tlower: %d\nupper: %d\n", ntohl(dtop.lower), ntohl(dtop.upper));
	}

	memset(temp, 0, sizeof(temp));
	if (imon2_get_exposure_property(fd, (struct ExposureProperty *)temp)) {
		printf("error for imon2_get_exposure_property\n");
	} else {
		over64((char *)&ep.lowertime, temp);
		over64((char *)&ep.uppertime, temp+8);
		over64((char *)&ep.timestep, temp+16);
		over64((char *)&ep.lowercycle, temp+24);
		over64((char *)&ep.uppercycle, temp+32);
		over64((char *)&ep.cyclestep, temp+40);

		out_info(is_out, "Exposure Property:\n\tlowertime=%e\n\tuppertime=%e\n\ttimestep=%e"
				"\n\tlowercycle=%e\n\tuppercycle=%e\n\tcyclestep=%e\n",
				ep.lowertime, ep.uppertime, ep.timestep, ep.lowercycle, ep.uppercycle, ep.cyclestep);
	}

	if (imon2_get_gain_property(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_gain_property\n");
	} else {
		out_info(is_out, "Gain Property:\n\t %d\n", ntohl(temp_long));
	}

	if (imon2_get_adoffset_property(fd, &ap)) {
		printf("error from imon2_get_adoffset_property\n");
	} else {
		out_info(is_out, "AdOffset Property:\n\tlower: %d\n\tupper: %d\n", ntohl(ap.lower), ntohl(ap.upper));
	}

	memset(temp, 0, sizeof(temp));
	if (imon2_get_cooling_property(fd, (struct CoolingProperty *)temp)) {
		printf("error for imon2_get_cooling_property\n");
	} else {
		over64((char *)&cp.lower, temp);
		over64((char *)&cp.upper, temp+8);
		over64((char *)&cp.step, temp+16);
		out_info(is_out, "Cooling Property:\n\tlower: %e\n\tupper: %e\n\tstep: %e\n",
				cp.lower, cp.upper, cp.step);
	}

	if (imon2_get_wavelength_property(fd, &wp)) {
		printf("error for imon2_get_wavelength_property\n");
	} else {
		out_info(is_out, "Wavelength Property:\n\tlowerwavelength: %d\n\tupperwavelength: %d\n", 
				ntohl(wp.lowerwavelength), ntohl(wp.upperwavelength));
	}

	if (imon2_get_imagesize(fd, &is)) {
		printf("error for imon2_get_imagesize\n"); 
	} else {
		out_info(is_out, "Image Size:\n\tsizeX: %d\n\tsizeY: %d\n", ntohl(is.sizeX), ntohl(is.sizeY));
	}

	memset(temp, 0, sizeof(temp));
	if (imon2_get_modulereadout_time(fd, (double *)temp)) {
		printf("error for imon2_get_modulereadout_time\n");
	} else {
		over64((char *)&temp_double, temp);
		out_info(is_out, "Module Readout Time:\n\t %e\n", temp_double);
	}
/*
 * set capture mode
 */
	if (imon2_set_capturemode(fd, htons(capture_mode))) {
		printf("error for imon2_set_capturemode\n");
	}

/*
 * get capture mode
 */
	if (imon2_get_capturemode(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_capturemode\n");
	} else {
		out_info(is_out, "Capture Mode:\n\t %d\n", ntohl(temp_long));
	}
/*
 * set data position
 */
	if (imon2_set_dataposition(fd, htonl(dataposition))) {
		printf("error for imon2_set_dataposition\n");
	}
/* 
 * get data position
 */
	if (imon2_get_dataposition(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_dataposition\n");
	} else {
		out_info(is_out, "Data Position:\n\t %d\n", ntohl(temp_long));
	}
/*
 * set data count
 */
	if (imon2_set_datacount(fd, htonl(datacount))) {
		printf("error for imon2_set_datacount\n");
	}
/*
 * get data count
 */
	if (imon2_get_datacount(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_datacount\n");
	} else {
		out_info(is_out, "DataCount:\n\t %d\n", ntohl(temp_long));
	}
/*
 * set data transmit
 */
	if (imon2_set_datatransmit(fd, htonl(datatransmit))) {
		printf("error for imon2_set_datatransmit\n");
	}
/*
 * get data transmit
 */
	if (imon2_get_datatransmit(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_datatransmit\n");
	} else {
		out_info(is_out, "DataTransmit:\n\t %d\n", ntohl(temp_long));
	}
	datatransmit=ntohl(temp_long);
/*
 * set data triggeroffset
 */
	if (imon2_set_datatrigger_offset(fd, htonl(triggeroffset))) {
		printf("error for imon2_set_datatrigger_offset\n");
	}
/*
 * get data triggeroffset
 */
	if (imon2_get_datatrigger_offset(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_datatrigger_offset\n");
	} else {
		out_info(is_out, "DataTriggerOffset:\n\t %d\n", ntohl(temp_long));
	}
/*
 * set exposure time
 */
	over64((char *)&temp_double, (char *)&exposure_time);
	if (imon2_set_exposuretime(fd, temp_double)) {
		printf("error for imon2_set_exposuretime\n");
	}
/*
 * get exposure time
 */
	memset(temp, 0, sizeof(temp));
	if (imon2_get_exposuretime(fd, (double *)temp)) {
		printf("error for imon2_get_exposuretime\n");
	} else {
		over64((char *)&temp_double, temp);
		out_info(is_out, "Exposure Time:\n\t %e\n", temp_double);
	}
/*
 * set exposure cycle
 */
	over64((char *)&temp_double, (char *)&exposure_cycle);
	if (imon2_set_exposurecycle(fd, temp_double)) {
		printf("error for imon2_set_exposurecycle\n");
	}
/*
 * get exposure cycle
 */
	memset(temp, 0, sizeof(temp));
	if (imon2_get_exposurecycle(fd, (double *)temp)) {
		printf("error for imon2_get_exposurecycle\n");
	} else {
		over64((char *)&temp_double, temp);
		out_info(is_out, "Exposure Cycle:\n\t %e\n", temp_double);
	}
/*
 * set trigger mode
 */
	if (imon2_set_triggermode(fd, htonl(trigger_mode))) {
		printf("error for imon2_set_triggermode\n");
	}
/*
 * get trigger mode
 */
	if (imon2_get_triggermode(fd, (__s32 *)&temp_long)) {
		out_info(is_out,"error for imon2_get_triggermode\n");
	} else {
		out_info(is_out,"Trigger Mode:\n\t %d\n", ntohl(temp_long));
	}
/*
 * set trigger polarity
 */
	if (imon2_set_triggerpolarity(fd, htonl(trigger_polarity))) {
		out_info(is_out,"error for imon2_set_triggerpolarity\n");
	}
/*
 * get trigger polarity
 */
	if (imon2_get_triggerpolarity(fd, (__s32 *)&temp_long)) {
		out_info(is_out,"error for imon2_get_triggerpolarity\n");
	} else {
		out_info(is_out,"Trigger Polarity:\n\t %d\n", ntohl(temp_long));
	}
/*
 * set trigger output
 */
	if (imon2_set_triggeroutput(fd, htonl(trigger_output))) {
		out_info(is_out,"error for imon2_set_triggeroutput\n");
	}
/*
 * get trigger output
 */
	if (imon2_get_triggeroutput(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_triggeroutput\n");
	} else {
		out_info(is_out, "Trigger Output:\n\t %d\n", ntohl(temp_long));
	}
/*
 * set gain
 */
	if (imon2_set_gain(fd, htonl(gain_mode), htonl(gain))) {
		printf("error for imon2_set_gain\n");
	}
/* 
 * get gain
 */
	if (imon2_get_gain(fd, gain_mode, (__s32 *)&temp_long)) {
		printf("error for imon2_get_gain\n");
	} else {
		out_info(is_out, "Gain for mode:\n\tmode: %d\n\tgain %d\n", gain_mode, ntohl(temp_long));
	}
/*
 * set adoffset
 */
	if (imon2_set_adoffset(fd, htonl(adoffset))) {
		printf("error for imon2_set_adoffset");
	}
/*
 * get adoffset
 */
	if (imon2_get_adoffset(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_adooffset\n");
	} else {
		out_info(is_out, "AdOffset:\n\t %d\n", ntohl(temp_long));
	}
/* get cooling temperature */
	if (imon2_get_coolingtemperature(fd, (double *)temp)) {
		printf("error for imon2_get_coolingtemperature\n");
	} else {
		over64((char *)&temp_double, temp);
		out_info(is_out, "Colling Temperature:\n\t %e\n", temp_double);
	}
/* get cooling status */
	if (imon2_get_coolingstatus(fd, (__s32 *)&temp_long)) {
		printf("error for imon2_get_coolingstatus\n");
	} else {
		out_info(is_out, "Cooling Status:\n\t %d\n", ntohl(temp_long));
	}
/* get current temperature */
	if (imon2_get_currenttemperature(fd, (double *)temp)) {
		printf("error for imon2_get_coolingtemperature\n");
	} else {
		over64((char *)&temp_double, temp);
		out_info(is_out, "Current Temperature:\n\t %e\n", temp_double);
	}
/* start capture in startcapturemode */
	out_info(is_out, "Index Start capture:	%i\n", startcaptmode);
	if (imon2_startcapture(fd, htons(startcaptmode))) {
		out_info(is_out,"error for start_capture\n");
		goto exit;
	}
/* get current time */
	timerclear(&tv_begin);
	timerclear(&tv_end);
	if (gettimeofday(&tv_begin, NULL)<0) {
		printf("error of gettimeofday\n");
		goto exit;
	}
	out_info(is_out, "Time of begining:\n\tsec: %lu\n\tu_sec: %lu\n", tv_begin.tv_sec, tv_begin.tv_usec);
/*
 * if out_dir is't NULL - writes all frames in the file
 */
	if (out_dir) {
		if ((fd_out=open(out_dir, O_WRONLY|O_APPEND|O_CREAT))<0) {
			qDebug("open out_dir error, fd: %i", fd_out);
		}
	}
			
	timerclear(&tv_begin);
	timerclear(&tv_end);
	if (gettimeofday(&tv_begin, NULL)<0) {
		printf("error of gettimeofday\n");
		goto exit;
	}
/* the main loop getting data from spectrometer */
	for (j=0; count_cycles ? j<count_cycles : 1; j++) {
		if (imon2_begin_cycle(fd, temp)) {
			out_info(is_out,"error for imon2_begin_cycle\n");
			goto exit;
		}
		/* get header */
		retval=read(fd, src_array, 0x200);
		if (retval<0) {
			printf("error for read header with errno %i\n", retval);
			goto exit;
		}
		/*
		 * out_frame_file(image_buf, 0x200);
		 */
		for (i=0; i<datatransmit; i++, count_frames++) {
			retval=read(fd, src_array, SIZE_DATA_ARRAY);
			wavelength=calc_wavelength(src_array, first_diff_array, second_diff_array,
				SIZE_DATA_ARRAY, &cc, this->different_level);
			if (retval<0) {
				qDebug("error for read image with errno");
				goto exit;
			}
			if (sprintf(doublestr, "%4.3f", wavelength)<0) {
				qDebug("error for sprintf");
				goto exit;
			}
			if (wavelength-prev_wavelength>this->change_level || prev_wavelength-wavelength>this->change_level) {
				prev_wavelength=wavelength;
				/*
				 * this need code for wavelength freq.
				 */
			}
			if (fd_out>0) 
				if (write(fd_out, src_array, SIZE_DATA_ARRAY)<0) {
					qDebug("write error for frame in file");
					break;
				}
			/*
			out_frame_file(image_buf, 0x400);	
			*/
		}
		if (imon2_end_cycle(fd, temp)) {
			qDebug("error for end cycle");
			goto exit;
		}
		if (gettimeofday(&tv_end, NULL)<0) {
			printf("error of gettimeofday\n");
			goto exit;
		}
		diff_time=DIFF_TIME_OF_DAY((&tv_begin), (&tv_end));
		poolHZ=(double)count_frames/(diff_time/1000000.0);
		if (sprintf(doublestr, "%3.1f", poolHZ)<0) {
			qDebug("error sprintf for poolHZ");
		} else {
			/*
			 * this is code for poolfreq. display
			*/
		}
		if (!is_run) break;
	}
/* end of main loop capture */
/* send "stop capture" to spectrometer */
	if (imon2_stopcapture(fd)) {
		qDebug("error for stop capture");
		goto exit;
	}
/* reading and sending needed ending packets to spcetrometer! */
	if (retval=imon2_begin_cycle(fd, temp)) { 
		if (retval<0) {
			qDebug("error for read header with errno %i\n", retval);
			goto exit;
		}
	}
	if (imon2_end_header(fd, temp)) {
		qDebug("error for imon2_end_header\n");
		goto exit;
	}
	if (imon2_end_cycle(fd, temp)) {
		qDebug("error for end cycle");
		goto exit;
	}
/* end all operations while sending */

exit:
	is_run=0;
	if (out_dir && fd_out>0) close(fd_out);
	return 0;
}


void MThread::run () {
	int fd;

	is_run=1;

	qDebug("exposure time: %f", exposure_time);
	qDebug("exposure cycle: %f", exposure_cycle);
	qDebug("gain: %i", gain);	
	qDebug("count cycles: %i", count_cycles);
	qDebug("data transmit: %i", data_transmit);
	qDebug("data count %i", data_count);
	qDebug("change level: %1.3f", change_level);
	qDebug("out file: %s", out_dir);
	qDebug("different level: %i", different_level);
	
	if ((fd=open(drv_name, O_RDONLY))<0) {
		qDebug("open error: ");
		goto exit;
	}
	if (start_capture (fd, exposure_time, exposure_cycle, is_out, gain,
				data_transmit, data_position, data_count, count_cycles,
				gain_mode, capture_mode, start_capture_mode, out_dir)) {
		qDebug("error for start_capture");
	}
	close(fd);
exit:
	qDebug("***** terminate run *****");
}

